/*

  rfbcounter
  
  This is a very simple VNC server which draws either a 
  fast-running counter or a clock display using 7-segment-style
  digits.  It does not have a framebuffer of its own.  The value stored
  in the counter is analagous to the pixels stored in a framebuffer -
  it is used to generate the VNC protocol on the fly when the client
  requests an update.

  Usage: 
  
    rfbcounter [-clock] display-number width [fg] [bg]

  eg. if you run

    rfbcounter 5 300

  you can then connect a VNC viewer to display 5 on that machine and you'll 
  get a counter 500 pixels wide. The fg & bg arguments are single byte values
  specifying foreground and background colours in BGR233. By prefixing with
  a zero, you can use octal to simplify color selection, for example 
  0007 0100 is bright red on a dark blue background.

  Note: This does not do any pixel translation.  Your viewer *must* cope
  with 8-bit bgr233 format for this to work.  On Unix, use the -bgr233 option
  to the viewer.

  This is not an example of good VNC code.  We ignore lots of messages we 
  shouldn't really ignore, for example.  It only uses the RRE encoding,
  because that's the easiest way to draw rectangles.

  Tristan Richardson & Quentin Stafford-Fraser
  For more info about VNC, see http://www.orl.co.uk/vnc
*/

// Set this for the endian of your machine. 0 = big, 1 = little 
static const int rfbEndianTest = 1;

#include <stdio.h>

#ifdef WIN32

#include <io.h>
#include "rfb.h"
#include <winsock.h>
#include <time.h>

#define sleep(x) Sleep((x)*1000)

#else

#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <unistd.h>
#include <X11/Xmd.h>
#include <sys/errno.h>

#endif

#include "rfbproto.h"
#include <sys/types.h>


int  ListenAtTcpAddr(unsigned short port);
void do_client(int sock);
int  draw_digit(int sock, int digit, int segl, int segw, 
		int fg, int bg, int xoffset, int yoffset);
int  draw_counter(int sock, int value, int numdigits, int segl, int segw, 
		  int fg, int bg);
void putSubrect(char* buf, int* lenp, int colour, 
		int x, int y, int w, int h);

int  ReadExact(int sock, char *buf, int len);
int  WriteExact(int sock, char *buf, int len);

int rfbMaxClientWait = 20000;	/* time (ms) after which we decide client has
				   gone away - needed to stop us hanging */

static const rfbPixelFormat BGR233Format = {
    8, 8, 1, 1, 7, 7, 3, 0, 3, 6
};

//
// Macro to compare pixel formats.
//

#define PF_EQ(x,y)							\
((x.bitsPerPixel == y.bitsPerPixel) &&				\
 (x.depth == y.depth) &&					\
 (x.trueColour == y.trueColour) &&				\
 ((x.bigEndian == y.bigEndian) || (x.bitsPerPixel == 8)) &&	\
 (!x.trueColour || ((x.redMax == y.redMax) &&			\
		    (x.greenMax == y.greenMax) &&		\
		    (x.blueMax == y.blueMax) &&			\
		    (x.redShift == y.redShift) &&		\
		    (x.greenShift == y.greenShift) &&		\
		    (x.blueShift == y.blueShift))))

//
// Macros for endian swapping.
//

#define Swap16(s) ((((s) & 0xff) << 8) | (((s) >> 8) & 0xff))

#define Swap32(l) (((l) >> 24) | \
		   (((l) & 0x00ff0000) >> 8)  | \
		   (((l) & 0x0000ff00) << 8)  | \
		   ((l) << 24))

#define Swap16IfLE(s) (*(const char *)&rfbEndianTest ? Swap16(s) : (s))

#define Swap32IfLE(l) (*(const char *)&rfbEndianTest ? Swap32(l) : (l))

int screenWidth, screenHeight;
int fg = 0;
int bg = 0xff;
int segmentWidth, segmentLength;
int numdigits = 4;
int segment_aspect = 3; /* length = 3 * width */
int digit_spacing = 2;  /* space is two segment widths */
int border;
int clockmode = 0;

#define PROGNAME "rfbcounter"

int main(int argc, char *argv[])
{
    int s, sock;
    unsigned short port;
    struct sockaddr addr;
    int addrlen;
    char *appname = argv[0];
	
    if (argc < 3) {
	fprintf(stderr,
		"usage: %s [-clock] server-number dispwidth [fg] [bg]\n",
		PROGNAME);
	exit(1);
    }
	
    if (strcmp(argv[1],"-clock") == 0) {
	clockmode = 1;
	argv++;
	argc--;
    }
	
    port = atoi(argv[1]);
    if (port < 100) port += 5900;
	
    screenWidth = atoi(argv[2]);
    border = screenWidth / 20;

    /* Each char is 2 segment widths + one length wide.
     * There are (numdigits-1) spaces between digits.
     */
    segmentWidth = (screenWidth-2 * border) / 
	( (numdigits * (2 + segment_aspect) + 
	   (numdigits - 1) * digit_spacing) );
    segmentLength = segmentWidth * segment_aspect;
    screenHeight = segmentWidth * (3+2*segment_aspect)  + 2 * border;
	
    if (argc >= 4)
	fg = strtol(argv[3], 0, 0);
	
    if (argc >= 5)
	bg = strtol(argv[4], 0, 0);
	
#ifdef WIN32
    {
        // Initialise winsock
        WORD wVersionRequested;
	WSADATA wsaData;
	wVersionRequested = MAKEWORD(2, 0);
	if (WSAStartup(wVersionRequested, &wsaData) != 0)
	    exit(1);
    }
#endif
	
    s = ListenAtTcpAddr(port);
    if (s < 0) {
	perror(PROGNAME);
	exit(1);
    }
	
    printf("%s: listening on port %d\n", appname, port);
	
    while (1) {
	int one = 1;
	rfbProtocolVersionMsg pv;
	CARD32 auth;
	rfbClientInitMsg cim;
	rfbServerInitMsg sim;
	char textbuf[1024];
		
        // Wait for an incoming connection
	addrlen = sizeof(addr);
	if ((sock = accept(s, &addr, &addrlen)) < 0) {
	    perror("accept");
	    exit(1);
	}
	
	// Disable Nagle's algorithm - improves performance 
	setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, 
		   (const char *) &one, sizeof(one));
		
	printf("%s: accepted connection\n", appname);
		
	// Send our protocol version & read the client's - we ignore it
	sprintf(pv,rfbProtocolVersionFormat,rfbProtocolMajorVersion,
		rfbProtocolMinorVersion);
    
	if (WriteExact(sock, pv, sz_rfbProtocolVersionMsg) < 0) {
	    perror("Writing protocol version");
	    close(sock);
	    exit(1);
	}
		
	if (ReadExact(sock, pv,	 sz_rfbProtocolVersionMsg) < 0) {
	    perror("Reading client protocol version");
	    close(sock);
	    exit(1);
	}
		
	// No authentication
	auth = Swap32IfLE(rfbNoAuth);
	if (WriteExact(sock, (char *) &auth, sizeof(auth)) < 0) {
	    perror("Writing NoAuth");
	    close(sock);
	    exit(1);
	}
		
	// Read & ignore client Init msg
	if (ReadExact(sock, (char *) &cim,	 sz_rfbClientInitMsg) < 0) {
	    perror("Reading client init msg");
	    close(sock);
	    exit(1);
	}
		
	// Server Init msg
	sim.framebufferWidth = Swap16IfLE(screenWidth);
	sim.framebufferHeight = Swap16IfLE(screenHeight);
	sim.format = BGR233Format;
	sim.format.redMax = Swap16IfLE(sim.format.redMax);
	sim.format.greenMax = Swap16IfLE(sim.format.greenMax);
	sim.format.blueMax = Swap16IfLE(sim.format.blueMax);
	sim.nameLength = strlen(PROGNAME);
	sim.nameLength = Swap32IfLE(sim.nameLength);
	strcpy(textbuf,PROGNAME);
	if (WriteExact(sock,(char *) &sim,	 sz_rfbServerInitMsg) < 0) {
	    perror("Sending server init msg");
	    close(sock);
	    exit(1);
	}
		
	if (WriteExact(sock, textbuf,strlen(PROGNAME)) < 0) {
	    perror("Sending server name");
	    close(sock);
	    exit(1);
	}
		
	// Then process this client
        do_client(sock);

	close(sock);
    }
    return 0;
}


void
do_client(int sock)
{
    while (1) {
	int n;
	CARD8 msgType;

	// Peek the first byte of the message to find its type
	if ((n = recv(sock, (char *)&msgType, 1, MSG_PEEK)) <= 0) {
	    if (n == 0)
		fprintf(stderr,"client gone\n");
	    else
		perror("recv");
	    return;
	}
		
	switch (msgType) {
			
	case rfbSetPixelFormat:
	    {
                // We read the client's pixel format, but in this simple 
                // program we are not going to do any translation, so if
                // the client doesn't either accept what we gave it, or 
                // request BGR233, we close connection.

		rfbSetPixelFormatMsg spf;
				
		if ((n = ReadExact(sock, (char *)&spf,
				   SIZEOF(rfbSetPixelFormatMsg))) <= 0) {
		    if (n == 0)
			fprintf(stderr,"rfbProcessClientMessage: client gone\n");
		    else
			perror("rfbProcessClientMessage: read");
		    return;
		}
				
		spf.format.bitsPerPixel = spf.format.bitsPerPixel;
		spf.format.depth = spf.format.depth;
		spf.format.trueColour = (spf.format.trueColour ? 1 : 0);
		spf.format.bigEndian = (spf.format.bigEndian ? 1 : 0);
		spf.format.redMax = Swap16IfLE(spf.format.redMax);
		spf.format.greenMax = Swap16IfLE(spf.format.greenMax);
		spf.format.blueMax = Swap16IfLE(spf.format.blueMax);
		spf.format.redShift = spf.format.redShift;
		spf.format.greenShift = spf.format.greenShift;
		spf.format.blueShift = spf.format.blueShift;
				
		if (!PF_EQ(spf.format, BGR233Format)) {
		    fprintf(stderr,"Client has wrong pixel format\n");
		    return;
		}
				
		break;
	    }

	case rfbFixColourMapEntries:
	    {
		fprintf(stderr,
			"rfbProcessClientMessage: cannot handle colourmaps\n");
		exit(1);
	    }


	case rfbSetEncodings:
            {
                
                // This reads but ignores the encoding list.
		// We simply check here that the client 
		// can do RRE encoding.

		rfbSetEncodingsMsg se;
		CARD32 enc;
		int i;
		int candoRRE = 0;
				
		if ((n = ReadExact(sock, (char *)&se,
				   SIZEOF(rfbSetEncodingsMsg))) <= 0) {
		    if (n == 0)
			fprintf(stderr,"rfbProcessClientMessage: client gone\n");
		    else
			perror("rfbProcessClientMessage: read");
		    return; 
		}
				
		se.nEncodings = Swap16IfLE(se.nEncodings);
		for (i=0; i<se.nEncodings; i++) {
		    if ((n = ReadExact(sock, 
				       (char *)&enc, sizeof(CARD32))) <= 0) {
			if (n == 0)
			    fprintf(stderr,
				    "rfbProcessClientMessage: client gone\n");
			else
			    perror("rfbProcessClientMessage: read");
			return;
		    }
		    if (Swap32IfLE(enc) == rfbEncodingRRE)
			candoRRE = 1;

		}
			
		if (!candoRRE) {
		    fprintf(stderr,"Client can't do RRE\n");
		    return;
		}
		break;
	    }
			

	case rfbFramebufferUpdateRequest:
	    {

                // This is the important bit.
                // The client has requested an update, so we need to send it
                // the display.

		rfbFramebufferUpdateRequestMsg sur;
		static counterval = 0;
		static oldcounterval = 0;
				
		if ((n = ReadExact(sock, 
			       (char *)&sur,
			       SIZEOF(rfbFramebufferUpdateRequestMsg)))<=0) {
		    if (n == 0)
		      fprintf(stderr,"rfbProcessClientMessage: client gone\n");
		    else
		      perror("rfbProcessClientMessage: read");
		    return;
		}
				
		if (sur.incremental) {
					
		    if (clockmode) {
			time_t t;
			struct tm *ptm;
			while (1) {
			    time(&t);
			    ptm = localtime(&t);
			    counterval = 100 * ptm->tm_hour + ptm->tm_min;
			    if (counterval != oldcounterval)
				break;
			    sleep(60-ptm->tm_sec);
			}
			oldcounterval = counterval;
		    } else {
			counterval = (counterval + 1) % 10000;
		    }
		    if (!draw_counter(sock, counterval, numdigits, 
				      segmentWidth, segmentLength, fg, bg))
			return;

		} else { /* blank the screen initially */

		    rfbFramebufferUpdateMsg su;
		    rfbFramebufferUpdateRectHeader rect;
		    rfbRREHeader hdr;
		    CARD8 colour;
		    su.type   = rfbFramebufferUpdate;
		    su.nRects = Swap16IfLE(1);
		    WriteExact(sock, (char *) &su, sz_rfbFramebufferUpdateMsg);
		    rect.r.x = Swap16IfLE(0);
		    rect.r.y = Swap16IfLE(0);
		    rect.r.w = Swap16IfLE(screenWidth);
		    rect.r.h = Swap16IfLE(screenHeight);
		    rect.encoding = Swap32IfLE(rfbEncodingRRE);
		    WriteExact(sock,(char *) &rect, 
			       sz_rfbFramebufferUpdateRectHeader);
		    hdr.nSubrects = Swap32IfLE(0);
		    WriteExact(sock,(char *) &hdr, sz_rfbRREHeader);
		    colour = bg;
		    WriteExact(sock,&colour, 1);
		    oldcounterval = 0;

		}
				
				
		break;
	    }
			
	case rfbKeyEvent:
	    {
                // Read and ignore key events, except for Q, which quits

		rfbKeyEventMsg key;
		if ((n = ReadExact(sock, (char *)&key,
				   SIZEOF(rfbKeyEventMsg))) <= 0) {
		    if (n == 0)
			fprintf(stderr,"rfbProcessClientMessage: client gone\n");
		    else
			perror("rfbProcessClientMessage: read");
		    return;
		}
		key.key = Swap32IfLE(key.key);
		key.key = key.key & 0xff;
		if ((key.key == 'q') || (key.key == 'Q'))
		    return;
		break;
	    }
			
	case rfbPointerEvent:
	    {
                // Read and ignore pointer events

		rfbPointerEventMsg ptr;
		if ((n = ReadExact(sock, (char *)&ptr,
				   SIZEOF(rfbPointerEventMsg))) <= 0) {
		    if (n == 0)
			fprintf(stderr,"rfbProcessClientMessage: client gone\n");
		    else
			perror("rfbProcessClientMessage: read");
		    return;
		}
		break;
	    }
			
	case rfbClientCutText:
	    {
                // Read and ignore clipboard messages

		rfbClientCutTextMsg ct;
		char *buf;
				
		if ((n = ReadExact(sock, (char *)&ct,
				   SIZEOF(rfbClientCutTextMsg))) <= 0) {
		    if (n == 0)
			fprintf(stderr,"rfbProcessClientMessage: client gone\n");
		    else
			perror("rfbProcessClientMessage: read");
		    return;
		}
				
		ct.length = Swap32IfLE(ct.length);
		buf = (char *) malloc(ct.length);
		if (buf == NULL) {
		    fprintf(stderr,"Cannot malloc %dl bytes\n", ct.length);
		    return;
		}
				
		if ((n = ReadExact(sock, buf,
				   ct.length)) <= 0) {
		    if (n == 0)
			fprintf(stderr,"rfbProcessClientMessage: client gone\n");
		    else
			perror("rfbProcessClientMessage: read");
		    return;
		}
				
		free(buf);
		break;
	    }
			
			
	default:
	    fprintf(stderr,
		    "rfbProcessClientMessage: unknown message type %d\n",
		    msgType);
	    fprintf(stderr,
		    "rfbProcessClientMessage: ... closing connection\n");
	    return;
	}	
    }
}


/*
 * ListenAtTcpAddr starts listening at the given TCP port.
 */

#define DO(x) if ((x) < 0) { close(sock); return -1; }

int ListenAtTcpAddr(port)
     unsigned short port;
{
    int sock;
    struct sockaddr_in addr;
    int one = 1;
	
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = INADDR_ANY;
	
    DO(sock = socket(AF_INET, SOCK_STREAM, 0));
    DO(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (const char *) &one, sizeof(one)));
    DO(bind(sock, (struct sockaddr *)&addr, sizeof(addr)));
    DO(listen(sock, 5));
    return sock;
}


/* Set the segments that make up each digit */

#define bottom      1
#define bottomLeft  2
#define bottomRight 4
#define middle      8
#define topLeft    16
#define topRight   32
#define top        64
#define all       127

CARD8 digits[10] = { 
    all&~middle,                            /* 0 */  
    bottomRight|topRight,                   /* 1 */
    all&~(topLeft|bottomRight),             /* 2 */
    all&~(topLeft|bottomLeft),              /* 3 */
    topLeft|middle|topRight|bottomRight,    /* 4 */
    all&~(topRight|bottomLeft),             /* 5 */
    all&~topRight,                          /* 6 */
    top|topRight|bottomRight,               /* 7 */
    all,                                    /* 8 */
    all&~bottomLeft                         /* 9 */
};

int draw_counter(int sock, int value, int numdigits, int segl, int segw, int fg, int bg) {
    rfbFramebufferUpdateMsg su;
    int xoffset = border;
    int yoffset = border;
    int i;

    // This one update will consist of a separate rectangle for each digit
    // Each rectangle uses RRE encoding, and the subrects draw the segments

    su.type = rfbFramebufferUpdate;
    su.nRects = Swap16IfLE(numdigits);
    WriteExact(sock, (char *) &su, sz_rfbFramebufferUpdateMsg);

    // We draw the digits right to left
    // The rightmost gopes here:
    xoffset += (numdigits-1) * (segl + 2 * segw + digit_spacing );

    for (i = 0; i < numdigits; i++) {
	if (!draw_digit(sock, value % 10, segl, segw, 
			fg, bg, xoffset, yoffset))
	    return 0;
	xoffset -= segl * ( (2 + segment_aspect) + digit_spacing );
        value = value / 10;
    }

    return 1;
}

int draw_digit(sock, digit, segl, segw, fg, bg, xoffset, yoffset)
     int sock;
     int digit;
     int segl, segw, fg, bg, xoffset, yoffset;
{
    rfbFramebufferUpdateRectHeader rect;
    rfbRREHeader hdr;
    char buf[256];
    int len = 0;
    int bits = digits[digit];
    char *rreheaderbuf;
	
    // A digit is made up of segments as follows:
    //     -      
    //    | |     where the length of each seg is segl
    //     -      and the width is segw.
    //    | |
    //     -

    // First calculate the overall dimensions of the segment
    rect.r.x = Swap16IfLE(xoffset);
    rect.r.y = Swap16IfLE(yoffset);
    rect.r.w = Swap16IfLE(2 * segl + segw);
    rect.r.h = Swap16IfLE(3 * segl + 2 * segw);
    rect.encoding = Swap32IfLE(rfbEncodingRRE);
	
    // Put this in the buffer ready to send
    memcpy(&buf[len], (char *)&rect,
	   SIZEOF(rfbFramebufferUpdateRectHeader));
    len += SIZEOF(rfbFramebufferUpdateRectHeader);
	
    rreheaderbuf = &buf[len];
    len += SIZEOF(rfbRREHeader);
	
    buf[len++] = bg;
	
    // Now insert in the buffer the commands needed to
    // draw each lit segment.
    hdr.nSubrects = 0;
	
    if (bits & bottom) {
	putSubrect(buf, &len, fg, segl, 2*segl+2*segw, segw, segl);
	hdr.nSubrects++;
    }
    if (bits & bottomLeft) {
	putSubrect(buf, &len, fg, 0, 2*segl+segw, segl, segw);
	hdr.nSubrects++;
    }
    if (bits & bottomRight) {
	putSubrect(buf, &len, fg, segl+segw, 2*segl+segw, segl, segw);
	hdr.nSubrects++;
    }
    if (bits & middle) {
	putSubrect(buf, &len, fg, segl, segl+segw, segw, segl);
	hdr.nSubrects++;
    }
    if (bits & topLeft) {
	putSubrect(buf, &len, fg, 0, segl, segl, segw);
	hdr.nSubrects++;
    }
    if (bits & topRight) {
	putSubrect(buf, &len, fg, segl+segw, segl, segl, segw);
	hdr.nSubrects++;
    }
    if (bits & top) {
	putSubrect(buf, &len, fg, segl, 0, segw, segl);
	hdr.nSubrects++;
    }
	
    hdr.nSubrects = Swap32IfLE(hdr.nSubrects);
    memcpy(rreheaderbuf, (char *)&hdr, SIZEOF(rfbRREHeader));
	
    // And then send the buffer to the viewer

    if (WriteExact(sock, buf, len) < 0) {
	perror("draw_digit: write");
	return 0;
    }
	
    return 1;
}

void putSubrect(char* buf, int* lenp, int colour, int x, int y, int w, int h)
{
    rfbRectangle r;
	
    buf[(*lenp)++] = colour;
    r.x = Swap16IfLE(x);
    r.y = Swap16IfLE(y);
    r.w = Swap16IfLE(w);
    r.h = Swap16IfLE(h);
    memcpy(&buf[*lenp], (char *)&r, SIZEOF(rfbRectangle));
    *lenp += SIZEOF(rfbRectangle);
}



/*
* ReadExact reads an exact number of bytes on a TCP socket.  Returns 1 if
* those bytes have been read, 0 if the other end has closed, or -1 if an error
* occurred (errno is set to ETIMEDOUT if it timed out).
*/

int ReadExact(int sock, char *buf, int len)
{
    int n;
 
    while (len > 0) {
	n = recv(sock, buf, len, 0);
	if (n > 0) {
	    buf += n;
	    len -= n;
        } else {
            return n;
	}
    }
    return 1;
}

/*
* WriteExact writes an exact number of bytes on a TCP socket.  Returns 1 if
* those bytes have been written, or -1 if an error occurred (errno is set to
* ETIMEDOUT if it timed out).
*/

int
WriteExact(int sock, char *buf, int len)
{
    int n;
	
    while (len > 0) {
	n = send(sock, buf, len,0);
		
	if (n > 0) {
	    buf += n;
	    len -= n;
	} else if (n == 0) {
      	    fprintf(stderr,"WriteExact: write returned 0?\n");
	    exit(1);
        } else {
            return n;
	}
    }
    return 1;
}

